import contextlib
import logging
from smtplib import SMTPException
from typing import Any

from django_ratelimit.decorators import ratelimit

from django.contrib import messages
from django.contrib.auth.views import LoginView, PasswordResetConfirmView, PasswordResetView
from django.core.exceptions import NON_FIELD_ERRORS, ImproperlyConfigured, ValidationError
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.decorators import method_decorator
from django.utils.encoding import force_str
from django.utils.http import urlsafe_base64_decode
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from django.views.generic import FormView, View

from core.forms import LoginForm, PasswordResetForm, RegistrationForm
from core.models import PlatformUser, UserCommunicationChannel
from core.tokens import channel_activation_token

logger = logging.getLogger(__name__)


@method_decorator(ratelimit(key='ip', rate='5/m', method=ratelimit.UNSAFE, block=False), name='post')
@method_decorator(ratelimit(key='post:username', rate='5/m', method=ratelimit.UNSAFE, block=False), name='post')
class BaseLoginView(LoginView):
    form_class = LoginForm
    next_page = 'backoffice:index'
    with_theme_switcher = False

    @method_decorator(sensitive_post_parameters('password'))
    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self) -> dict[str, Any]:
        return {'with_theme_switcher': self.with_theme_switcher, **super().get_form_kwargs()}

    def form_invalid(self, form: LoginForm) -> HttpResponse:
        logger.info('Invalid login attempt for username %s logged from %s', form.cleaned_data.get("username"), self.request.META.get('REMOTE_ADDR'))
        return super().form_invalid(form)

    def form_valid(self, form: LoginForm):
        user = form.get_user()
        self.request.session['theme'] = form.cleaned_data.get('theme', user.theme)
        res = super().form_valid(form)
        logger.info('User %s logged in from %s', user.username, self.request.META.get('REMOTE_ADDR'))
        return res


class BasePasswordResetView(PasswordResetView):
    email_template_name = 'registration/core_password_reset_email.html'
    extra_email_context = None
    form_class = PasswordResetForm
    html_email_template_name = None
    subject_template_name = 'registration/core_password_reset_subject.txt'
    template_name = 'registration/password_reset.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.setdefault('retry', self.request.GET.get('retry', 'False') == 'True' if hasattr(self.request, 'GET') else False)
        return context


class BasePasswordResetConfirmView(PasswordResetConfirmView):
    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        resp = super().dispatch(*args, **kwargs)
        if hasattr(resp, 'context_data') and not resp.context_data['validlink']:
            return HttpResponseRedirect(f'{self.failure_url}?retry=True')
        return resp


@method_decorator(ratelimit(key='ip', rate='5/m', method=ratelimit.UNSAFE, block=False), name='post')
@method_decorator(ratelimit(key='post:username', rate='5/m', method=ratelimit.UNSAFE, block=False), name='post')
class BaseRegistrationView(FormView):
    email_template_name = 'registration/registration_email.html'
    extra_email_context = None
    form_class = RegistrationForm
    html_email_template_name = None
    subject_template_name = 'registration/registration_subject.txt'
    success_url = reverse_lazy('signup_done')
    activation_url = 'signup_activate'
    template_name = 'registration/registration_form.html'
    title = _('Registration')
    token_generator = channel_activation_token

    @method_decorator(sensitive_post_parameters('password1', 'password2'))
    def post(self, request: HttpRequest, *args: str, **kwargs: Any) -> HttpResponse:
        return super().post(request, *args, **kwargs)

    def get_form_kwargs(self) -> dict[str, Any]:
        return {**super().get_form_kwargs(), 'request': self.request}

    def form_valid(self, form):
        opts = {
            'activation_url': self.activation_url,
            'use_https': self.request.is_secure(),
            'token_generator': self.token_generator,
            'email_template_name': self.email_template_name,
            'subject_template_name': self.subject_template_name,
            'request': self.request,
            'html_email_template_name': self.html_email_template_name,
            'extra_email_context': self.extra_email_context,
        }
        try:
            form.save(**opts)
            logger.info('Sent account activation mail to "%(email)s" for %(user)s.', {'email': form.instance.email, 'user': form.instance.username})
            messages.success(self.request, _('Registration__email__sending__success'))
        except SMTPException:
            logger.info('Failed to send account activation mail to "%(email)s" for %(user)s.', {'email': form.instance.email, 'user': form.instance.username})
            messages.error(self.request, _('Registration__email__sending__failure'))
            form.errors[NON_FIELD_ERRORS] = ValidationError(_('Registration__email__sending__failure'))
            return self.form_invalid(form)
        return super().form_valid(form)


class BaseRegistrationActivationView(View):
    redirect_target_failure = None
    redirect_target_success = None

    logger = logging.getLogger(__name__)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.redirect_target_failure is None or self.redirect_target_success is None:
            raise ImproperlyConfigured('View is missing either a redirect_target_success or a redirect_target_failure.')

    def get(self, request, uid_b64, channel_id, token):
        user, channel = None, None  # ensure variables exists if any of the next two fail, otherwise the logger call fails
        try:
            uid = force_str(urlsafe_base64_decode(uid_b64))
            user = PlatformUser.objects.get(pk=uid)
            channel = user.communication_channels.get(id=channel_id)
        except PlatformUser.DoesNotExist:
            messages.warning(self.request, _('registration__activation__user_does_not_exist'))
            self.logger.warning(
                'Account activation failed (uid_b64=%(uid_b64)s, channel=%(channel_id)s, token=%(token)s): User does not exist',
                {
                    'uid_b64': uid_b64,
                    'channel_id': channel_id,
                    'token': token,
                },
            )
            return redirect(self.redirect_target_failure)
        except UserCommunicationChannel.DoesNotExist:
            messages.warning(self.request, _('registration__activation__invalid_code'))
            self.logger.warning(
                'Account activation failed (user=%(user)s, uid_b64=%(uid_b64)s, channel=%(channel_id)s, token=%(token)s): Channel does not exist',
                {
                    'user': user,
                    'uid_b64': uid_b64,
                    'channel_id': channel_id,
                    'token': token,
                },
            )
            return redirect(self.redirect_target_failure)

        if not channel_activation_token.check_token(channel, token):
            messages.warning(self.request, _('registration__activation__invalid_code'))
            self.logger.warning(
                'Account activation failed (user=%(user)s, uid_b64=%(uid_b64)s, channel=%(channel_id)s, token=%(token)s) due to invalid token.',
                {
                    'user': user,
                    'uid_b64': uid_b64,
                    'channel_id': channel_id,
                    'token': token,
                },
            )
            return redirect(self.redirect_target_failure)

        if channel.channel != UserCommunicationChannel.Channel.MAIL or user.email is None or user.email != channel.address:
            messages.warning(self.request, _('registration__activation__no_account'))
            self.logger.warning(
                'Account activation failed for non-registration-channel (user=%(user)s, uid_b64=%(uid_b64)s, channel=%(channel_id)s, token=%(token)s)',
                {
                    'user': user,
                    'uid_b64': uid_b64,
                    'channel_id': channel_id,
                    'token': token,
                },
            )
            return redirect(self.redirect_target_failure)

        # first time registration: user.email was still set to the mail communication channel's address
        user.is_active = True
        user.email = ''
        channel.is_verified = True
        user.save()
        channel.save()

        with contextlib.suppress(ImproperlyConfigured):
            if (hasattr(self, 'conference') and (conference := getattr(self, 'conference'))) or (
                hasattr(self, 'conf') and (conference := getattr(self, 'conf'))
            ):
                conference.users.get_or_create(user=user)

        messages.success(self.request, _('registration__activation__success'))
        self.logger.info('Account activated successfully: %s', user.username)

        return redirect(self.redirect_target_success)
